use std::{iter, num::NonZeroU8};

use anyhow::Result;
use itertools::Itertools;

use super::{
	util::{fill_partials, TreeCoords},
	MaximizerAgent, MinimizerAgent, RandomAgent,
};
use crate::{
	agents::util::{find_partials, free_chars},
	board::{Board, PlayerAction, TreeElement},
	util::Symb,
};

pub struct MinMaxTree {}

fn find_best_numbers_v1<'a, F>(
	tree: &TreeElement,
	partials: &[TreeCoords],
	numbers: impl Iterator<Item = &'a Symb>,
	minimize: bool,

	// Returns true if we want to maximize the given partial,
	// and false if we want to fix it.
	filter: F,
) -> Vec<Symb>
where
	F: Fn(&&TreeCoords) -> bool,
{
	// Fill maximizer slots with arbitrary numbers
	let min_tree_base = fill_partials(
		tree,
		partials.iter().filter(|x| !filter(x)),
		iter::repeat(&Symb::Number(NonZeroU8::new(5).unwrap())),
	);

	let partials_to_optimize: Vec<TreeCoords> = partials.iter().filter(filter).cloned().collect();
	let n_empty = free_chars(tree, partials_to_optimize.iter()).len();
	println!("{:?}", n_empty);

	let trees: Vec<(f32, Vec<&Symb>)> = numbers
		.permutations(n_empty)
		.unique()
		.filter_map(move |l| {
			let mut i = l.iter();
			let mut tmp_tree = min_tree_base.clone();
			for p in &partials_to_optimize {
				let x = p.get_from_mut(&mut tmp_tree).unwrap();

				let x_str = match x {
					TreeElement::Partial(s) => s,
					_ => unreachable!(),
				};
				let mut new_str = String::new();
				for c in x_str.chars() {
					if c == '_' {
						new_str.push_str(&format!("{}", i.next().unwrap()))
					} else {
						new_str.push(c);
					}
				}
				*x = TreeElement::Number(new_str.parse().unwrap())
			}

			println!("{:?}", tmp_tree);
			tmp_tree.evaluate().map(|x| (x, l))
		})
		.collect();

	let mut best_list: Option<Vec<&Symb>> = None;
	let mut best_value: Option<f32> = None;

	for (x, list) in trees {
		if let Some(m) = best_value {
			if (minimize && x < m) || (!minimize && x > m) {
				best_value = Some(x);
				best_list = Some(list);
			}
		} else {
			best_value = Some(x);
			best_list = Some(list);
		}
	}

	best_list.unwrap().into_iter().cloned().collect()
}

fn find_best_numbers(
	tree: &TreeElement,
	partials: &[TreeCoords],

	// The numbers we're allowed to add, sorted in ascending order
	available_numbers: &[Symb],
) -> TreeElement {
	// Fill all empty slots with fives
	let tree_filled = fill_partials(
		tree,
		partials.iter(),
		iter::repeat(&Symb::Number(NonZeroU8::new(5).unwrap())),
	);

	let base = tree_filled.evaluate().unwrap();

	// Test each slot:
	// Increase its value by 1, and record its effect on the
	// expression's total value.
	// This isn't a perfect metric, but it's pretty good.
	let mut slots: Vec<(usize, &TreeCoords, usize, f32)> = free_chars(tree, partials.iter())
		.into_iter()
		.enumerate()
		.map(|(i_slot, (c, i))| {
			let mut new_tree = tree_filled.clone();
			let p = c.get_from_mut(&mut new_tree).unwrap();
			match p {
				TreeElement::Partial(s) => s.replace_range(i..i + 1, "6"),
				_ => unreachable!(),
			}
			// This shouldn't ever be None.
			(i_slot, c, i, new_tree.evaluate().unwrap() - base)
		})
		.collect();

	// Sort by least to most influence
	slots.sort_by(|a, b| a.3.partial_cmp(&b.3).unwrap());

	let all_symbols = {
		// We need this many from the bottom, and this many from the top.
		let neg_count = slots.iter().filter(|(_, _, _, x)| *x <= 0.0).count();
		let pos_count = slots.iter().filter(|(_, _, _, x)| *x > 0.0).count();

		let mut a_iter = available_numbers
			.iter()
			.take(neg_count)
			.chain(available_numbers.iter().rev().take(pos_count).rev());

		let mut g = slots
			// Group slots with equal weights
			// and count the number of elements in each group
			.iter()
			.group_by(|x| x.3)
			.into_iter()
			.map(|(_, x)| x.count())
			// Generate the digits we should try for each group of
			// equal-weight slots
			.map(|s| {
				(0..s)
					.map(|_| a_iter.next().unwrap().clone())
					.permutations(s)
					.unique()
					.collect_vec()
			})
			// Now, covert this to an array of all cartesian products
			// of this set of sets
			.multi_cartesian_product()
			.map(|x| x.iter().flatten().cloned().collect_vec())
			.map(|v| slots.iter().zip(v).collect_vec())
			.collect_vec();

		// Sort these vectors so the order of values
		// matches the order of empty slots
		g.iter_mut()
			.for_each(|v| v.sort_by(|(a, _), (b, _)| a.0.partial_cmp(&b.0).unwrap()));
		g.into_iter()
			.map(|v| v.into_iter().map(|(_, s)| s).collect_vec())
	};

	let mut best_tree = None;
	let mut best_value = None;
	for i in all_symbols {
		let tmp_tree = fill_partials(&tree, partials.iter(), i.iter());
		let val = tmp_tree.evaluate();

		if let Some(val) = val {
			if let Some(best) = best_value {
				if val > best {
					best_value = Some(val);
					best_tree = Some(tmp_tree)
				}
			} else {
				best_value = Some(val);
				best_tree = Some(tmp_tree)
			}
		}
	}

	best_tree.unwrap()
}

impl MinMaxTree {}

impl MinimizerAgent for MinMaxTree {
	fn step_min(&mut self, board: &Board) -> Result<PlayerAction> {
		let tree = board.to_tree();
		let partials = find_partials(&tree);

		let available_numbers = (0..=9)
			.map(|x| match x {
				0 => Symb::Zero,
				x => Symb::Number(NonZeroU8::new(x).unwrap()),
			})
			.filter(|x| !board.contains(*x))
			.collect::<Vec<_>>();

		// For the code below, we must guarantee that
		// that is, min_slots + max_slots <= available_numbers.len
		if available_numbers.len() < free_chars(&tree, partials.iter()).len() {
			return RandomAgent {}.step_max(board);
		}

		let t = find_best_numbers(&tree, &partials, &available_numbers);

		println!("{:?}", t);
		RandomAgent {}.step_max(board)
	}
}
